#ifndef ZAD2_VARIABLE_H
#define ZAD2_VARIABLE_H

#include <utility>
#include <map>

#include "../TypeDefs.h"
#include "../Info.h"

class Variable : std::enable_shared_from_this<Variable> {
public:
	Variable() = default;
	explicit Variable(VarInfoPtr i) : info(std::move(i)), type(info->type) {};
	explicit Variable(int constVal) : type(new Int), constExpr(true), val(constVal) {};
	explicit Variable(std::string symbolName) : name(std::move(symbolName)) {};
	VarInfoPtr info;
	TypePtr type;
	bool constExpr{false};
	int val{0};
	std::string name;
	int registerColor = 0;

	vector<QuadruplePtr> uses;
	vector<QuadruplePtr> writes;

	int localOffset = 0;

	void clear(const QuadruplePtr& q) {
		uses.erase(std::remove(uses.begin(), uses.end(), q), uses.end());
		writes.erase(std::remove(writes.begin(), writes.end(), q), writes.end());
	}
};

class VariableLocation {
public:
	VariableLocation() = default;
	VariableLocation(VariablePtr base, VariablePtr index, int multiplier, int offset)
		: base(std::move(base)), index(std::move(index)), multiplier(multiplier), offset(offset)
	{}
	VariablePtr base;
	VariablePtr index;
	int multiplier{};
	int offset{};

	string toString() const {
		std::string ret = "[" + base->name;
		if (index) if (multiplier || !(index->constExpr || !index->val)) ret += " + " + to_string(multiplier) + "*" + index->name;
		if (offset) ret += " + " + to_string(offset);
		return ret + "]";
	}
};


class VariableLayout {
public:
	map<VarInfoPtr, pair<VariablePtr, VariablePtr>> changes;
	void add(const VarInfoPtr& info) {
		if (!changes.count(info)) {
			changes[info] = {info->loc, nullptr};
		}
	}

	void capture() {
		for (auto &p : changes) {
			p.second.second = p.first->loc;
		}
	}

	void revert() {
		for (const auto& p : changes) {
			VarInfoPtr info = p.first;
			info->loc = p.second.first;
		}
	}
};

#endif //ZAD2_VARIABLE_H
